# # Analysis of The Triceps Skinfold Dataset
# # Section 6 of 
# N. Veraverbeke, I. Gijbels and M. Omelka  
# Pre-adjusted nonparametric estimation of a conditional distribution function, 
# JRSSB 


rm(list=ls());

library(locpol);

# Speed up computation when more cores of a processor are available
library(doMC);
n.cores <- 1; # number of available cores of the processor 
registerDoMC(cores=n.cores);


# # READING THE DATA SET:  
# Data are available in different formats and 
# can be downloaded from the address: 
# http://portal.uni-freiburg.de/imbi/Royston-Sauerbrei-book/index.html#datasets
# or directly 
# http://portal.uni-freiburg.de/imbi/Royston-Sauerbrei-book/Multivariable_Model-building/downloads/datasets/triceps.zip
# In this script the file 'triceps.csv' is used. 

DATA <- read.table("triceps.csv", sep=",", header=T);

# # Conditional distribution function estimation 
cond.cdf <- function(x, y, x1, y1, logtransf = FALSE){
# x ... observed X-values
# y ... observed Y-values
# x1 ... values of x, for which F_{x}(y) is estimated 
# y1 ... values of y, for which F_{x}(y) is estimated 
  
# Used bandwidths   
  BW2 <- 0.5;  # b_{n} in (20) of Remark 2
  BW2.het <- 0.5; # b_{n} in (20) of Remark 2
  
#   x <- DATA$age; 
  
  
#  log-transformation of response y?
  if(logtransf == TRUE){
    y <- log(y); 
  }
#   else{
#     y <- DATA$triceps; 
#   }
  
#   Sorting the observations with respect to x (age)
  qqq <- order(x);
  x <- x[qqq];
  y <- y[qqq];
  
  n <- nrow(DATA);

  y.kvant <- y1;
  ny.kvant <- length(y.kvant);
  
  x.kvant <- x1;
  nx.kvant <- length(x.kvant);
  
#   Integral of Epanechnikov kernel 
  K <-  function(x){
    res <- 3/4*(x - x^3/3 + 2/3);
    res[x > 1] <- 1;
    res[x < -1] <- 0;
      
    return(res);
  }
  
#   Bi-weight kernel 
  kkk <- biweigK;
  
  
  bw.max <- max(x)-min(x);
  skala <- 3*IQR(x);
  xs <- x/skala;
  
# Bandwidth to adjust the response through a location-scale model   
  bw0 <- regCVBwSelC(xs, y, kernel=kkk, deg=1)
  
#   A minimal bandwidth so that at least 100-observations are 
#   used at each of the observed points. 
  h1.min <- max(apply(abs(outer(x,x,"-")), 1, quantile, probs=100/n));
  h1 <- max(n^0.1*skala*bw0, h1.min);
  
#   Upper bound for the bandwidth 
  h2.kvant0 <- apply(abs(outer(c(x.kvant),x,"-")), 1, quantile, probs=250/n);
  h.kvant <- pmin(fitted(loess(h2.kvant0~c(x.kvant))), max(h2.kvant0));
  
  x0 <- c(x, x.kvant);  
  W <- locLinWeightsC(x, xeval=x0, bw=h1, kernel=kkk)$locWeig;  
  
  muhat <- W%*%y;
  ehat.res <- y - muhat[1:n]; # homoscedastic mean-model  
  
#   Determining the bandwidth for nonparametric scale estimation 
  res2 <- ehat.res^2;  
  bw0.scale <- try(regCVBwSelC(xs, res2, kernel=kkk, deg=1));
    
  BW.scale <- min(max(n^0.1*skala*bw0.scale, h1.min), bw.max);
    
  W <- locLinWeightsC(x, xeval=c(x,x.kvant), bw=BW.scale, kernel=kkk)$locWeig;
  
#  Estimate of the conditional variance  
  temp.sigma2 <- W%*%ehat.res^2;
  
  
#   If the estimate of the conditional variance to low, 
#   try to use the NW-weights 
  if(min(temp.sigma2) < 1e-6){
      bw0.scale <- try(regCVBwSelC(xs, res2, kernel=kkk, deg=0));
    
      BW.scale <- min(max(n^0.1*skala*bw0.scale, h1.min), bw.max);
    
      W <- locCteWeightsC(x, xeval=c(x,x.kvant), bw=BW.scale, kernel=kkk)$locWeig;
      
      temp.sigma2 <- W%*%res2;
    }
  
  
  sigma2 <- temp.sigma2[1:n];
  
#   Estimate of the scale function (conditional sd)
  sigmahat <- sqrt(sigma2);
  
  mu.x <- muhat[(n+1):(n+length(x.kvant))];
  sigma2.x <- temp.sigma2[(n+1):(n+length(x.kvant))];
  sigma.x <- sqrt(sigma2.x); 
   
  ehat.res2 <- ehat.res/sigmahat;   # heteroscedastic mean-model
    
#     Conditional distribution estimation 
  Yy <- outer(y, y.kvant, "-");
  Yy.lp1 <- outer(ehat.res, y.kvant, "-");
  
#   If more than one core of the processor are available 
  if(n.cores > 1){
# #     Multiple cores computation
    q <- foreach(k = 1:nx.kvant, .combine="rbind", .inorder=TRUE) %dopar% {
      wn <- locLinWeightsC(x, xeval=x.kvant[k], bw=h.kvant[k], kernel=EpaK)$locWeig;
      
      wn[wn < 0] <- 0;
      wn <- wn/sum(wn);
      
#       browser();
      Yy.K <- K((-1)*Yy/(BW2*sigma.x[k]));
      Yy.lp1.K <- K((-mu.x[k]-Yy.lp1)/(BW2*sigma.x[k]));
      Yy.lp2 <- outer(ehat.res2, (y.kvant-mu.x[k])/sigma.x[k], "-");
      Yy.lp2.K <- K((-1)*Yy.lp2/BW2.het);
#       Yy.lp2 <- K((-1)*Yy.lp2/0.5);

      # standard estimate of conditional cdf
      estim.stand <- wn%*%Yy.K;    
      # estimate of conditional cdf adjusted through a nonparametric location model 
      estim.lp1 <- wn%*%Yy.lp1.K;
      # estimate of conditional cdf adjusted through a nonparametric location-scale model model 
      estim.lp2 <- wn%*%Yy.lp2.K;
      
      c(estim.lp1, estim.lp2, estim.stand);
    }
    
    estim.lp1 <- t(q[,1:ny.kvant]);
    estim.lp2 <- t(q[,(ny.kvant+1):(2*ny.kvant)]);
    estim.stand <- t(q[,(2*ny.kvant+1):(3*ny.kvant)]);
#     
  }else{  
#     browser();
    estim.stand <- matrix(NA, nrow=ny.kvant, ncol=nx.kvant);  
    estim.lp1 <- matrix(NA, nrow=ny.kvant, ncol=nx.kvant);  
    estim.lp2 <- matrix(NA, nrow=ny.kvant, ncol=nx.kvant); 
    
# # Standard loop for one-core computation    
    for(k in 1:length(x.kvant)){
      wn <- locLinWeightsC(x, xeval=x.kvant[k], bw=h.kvant[k], kernel=EpaK)$locWeig;
      
      Yy.K <- K((-1)*Yy/(BW2*sigma.x[k]));
      Yy.lp1.K <- K((-mu.x[k]-Yy.lp1)/BW2);      
      Yy.lp2 <- outer(ehat.res2, (y.kvant-mu.x[k])/sigma.x[k], "-");
      Yy.lp2 <- K((-1)*Yy.lp2/BW2.het);

      estim.stand[,k] <- wn%*%Yy.K;
      estim.lp1[,k] <- wn%*%Yy.lp1.K;
      estim.lp2[,k] <- wn%*%Yy.lp2;
    }     
  }
 
# #  Preparing results 
  results <- list();
  
  # standard estimate of a conditional cdf
  results$stand <- estim.stand;
  # estimate of a conditional cdf adjusted through a nonparametric location model 
  results$lp1 <- estim.lp1;
  # estimate of a conditional cdf adjusted through a nonparametric location-scale model model 
  results$lp2 <- estim.lp2;

  results$x1 <- x1;
  results$y1 <- y1;

  
#   browser();
  
  invisible(results);
}

# Ages to estimate the conditional cdf F_{x}(y)
x0 <- seq(min(DATA$age), max(DATA$age), by=0.5);
# y-points to estimate the conditional cdf F_{x}(y) 
y0 <- seq(min(DATA$triceps),31, by=0.1);
# ylog <- log(seq(min(DATA$triceps),31, by=0.1));
print(system.time(fit <- cond.cdf(x=DATA$age, y=DATA$triceps, x1=x0, y1=y0, logtransf=FALSE)));


# # Figure 11 of the manuscript 
# postscript("triceps.eps", paper="special", width=16, height=8);
par(mfrow=c(1,2));
# Subfigure with data points 
plot(DATA$age, DATA$triceps, ylim=c(0,max(DATA$triceps)), xlab="Age (yr)", ylab="Triceps Skinfold", main="(a) Original data", cex=0.25, cex.main=2, cex.lab=1.45, cex.axis=1.45);
# 
# # Subfigure with estimated quantiles 
contour(x=x0, y= y0, z=t(fit$stand), levels=c(0.03,0.1,0.25,0.50,0.75,0.9,0.97), lty="solid", ylim=c(0,max(DATA$triceps)), xlab="Age (yr)", ylab="Triceps Skinfold", main="(b) Estimated quantiles", cex.main=2, cex.lab=1.45, cex.axis=1.45, labcex=1.1);
contour(x=x0, y= y0, z=t(fit$lp2), levels=c(0.03,0.1,0.25,0.50,0.75,0.9,0.97), add=TRUE, lty="dotted", drawlabels=FALSE, lwd=3);
xkv <- quantile(DATA$age, c(0.1,0.9));
abline(v=xkv[1], lty="dotted", lwd=2);
abline(v=xkv[2], lty="dotted", lwd=2);
legend(0, max(DATA$triceps), lty=c("solid", "dotted"), lwd=c(1,3), legend=c("stand", "lp-het"), cex=2, bg="white");
par(mfrow=c(1,1));
# dev.off();
